package org.apache.bcel.classfile;

import java.io.FilterWriter;
import java.io.FilterReader;
import java.util.Locale;
import java.io.PrintWriter;
import java.io.PrintStream;
import java.util.List;
import org.apache.commons.lang3.ArrayUtils;
import java.util.ArrayList;
import java.io.Writer;
import java.io.CharArrayWriter;
import java.io.OutputStream;
import java.util.zip.GZIPOutputStream;
import java.util.Arrays;
import java.io.InputStream;
import java.util.zip.GZIPInputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.Reader;
import java.io.CharArrayReader;
import java.io.IOException;
import org.apache.bcel.util.ByteSequence;
import org.apache.bcel.Const;

public abstract class Utility
{
    private static final ThreadLocal<Integer> CONSUMER_CHARS;
    private static boolean wide;
    private static final int FREE_CHARS = 48;
    private static final int[] CHAR_MAP;
    private static final int[] MAP_CHAR;
    private static final char ESCAPE_CHAR = '$';
    
    public static String accessToString(final int accessFlags) {
        return accessToString(accessFlags, false);
    }
    
    public static String accessToString(final int accessFlags, final boolean forClass) {
        final StringBuilder buf = new StringBuilder();
        int p = 0;
        int i = 0;
        while (p < 32768) {
            p = pow2(i);
            Label_0066: {
                if ((accessFlags & p) != 0x0) {
                    if (forClass) {
                        if (p == 32) {
                            break Label_0066;
                        }
                        if (p == 512) {
                            break Label_0066;
                        }
                    }
                    buf.append(Const.getAccessName(i)).append(" ");
                }
            }
            ++i;
        }
        return buf.toString().trim();
    }
    
    private static short byteToShort(final byte b) {
        return (b < 0) ? ((short)(256 + b)) : b;
    }
    
    public static String classOrInterface(final int accessFlags) {
        return ((accessFlags & 0x200) != 0x0) ? "interface" : "class";
    }
    
    public static int clearBit(final int flag, final int i) {
        final int bit = pow2(i);
        return ((flag & bit) == 0x0) ? flag : (flag ^ bit);
    }
    
    public static String codeToString(final byte[] code, final ConstantPool constantPool, final int index, final int length) {
        return codeToString(code, constantPool, index, length, true);
    }
    
    public static String codeToString(final byte[] code, final ConstantPool constantPool, final int index, final int length, final boolean verbose) {
        final StringBuilder buf = new StringBuilder(code.length * 20);
        try (ByteSequence stream = new ByteSequence(code)) {
            for (int i = 0; i < index; ++i) {
                codeToString(stream, constantPool, verbose);
            }
            int i = 0;
            while (stream.available() > 0) {
                if (length < 0 || i < length) {
                    final String indices = fillup(stream.getIndex() + ":", 6, true, ' ');
                    buf.append(indices).append(codeToString(stream, constantPool, verbose)).append('\n');
                }
                ++i;
            }
        }
        catch (final IOException e) {
            throw new ClassFormatException("Byte code error: " + buf.toString(), e);
        }
        return buf.toString();
    }
    
    public static String codeToString(final ByteSequence bytes, final ConstantPool constantPool) throws IOException {
        return codeToString(bytes, constantPool, true);
    }
    
    public static String codeToString(final ByteSequence bytes, final ConstantPool constantPool, final boolean verbose) throws IOException {
        final short opcode = (short)bytes.readUnsignedByte();
        int defaultOffset = 0;
        int noPadBytes = 0;
        final StringBuilder buf = new StringBuilder(Const.getOpcodeName(opcode));
        if (opcode == 170 || opcode == 171) {
            final int remainder = bytes.getIndex() % 4;
            noPadBytes = ((remainder == 0) ? 0 : (4 - remainder));
            for (int i = 0; i < noPadBytes; ++i) {
                final byte b;
                if ((b = bytes.readByte()) != 0) {
                    System.err.println("Warning: Padding byte != 0 in " + Const.getOpcodeName(opcode) + ":" + b);
                }
            }
            defaultOffset = bytes.readInt();
        }
        switch (opcode) {
            case 170: {
                final int low = bytes.readInt();
                final int high = bytes.readInt();
                final int offset = bytes.getIndex() - 12 - noPadBytes - 1;
                defaultOffset += offset;
                buf.append("\tdefault = ").append(defaultOffset).append(", low = ").append(low).append(", high = ").append(high).append("(");
                final int[] jumpTable = new int[high - low + 1];
                for (int j = 0; j < jumpTable.length; ++j) {
                    buf.append(jumpTable[j] = offset + bytes.readInt());
                    if (j < jumpTable.length - 1) {
                        buf.append(", ");
                    }
                }
                buf.append(")");
                break;
            }
            case 171: {
                final int npairs = bytes.readInt();
                final int offset = bytes.getIndex() - 8 - noPadBytes - 1;
                final int[] match = new int[npairs];
                final int[] jumpTable = new int[npairs];
                defaultOffset += offset;
                buf.append("\tdefault = ").append(defaultOffset).append(", npairs = ").append(npairs).append(" (");
                for (int j = 0; j < npairs; ++j) {
                    match[j] = bytes.readInt();
                    jumpTable[j] = offset + bytes.readInt();
                    buf.append("(").append(match[j]).append(", ").append(jumpTable[j]).append(")");
                    if (j < npairs - 1) {
                        buf.append(", ");
                    }
                }
                buf.append(")");
                break;
            }
            case 153:
            case 154:
            case 155:
            case 156:
            case 157:
            case 158:
            case 159:
            case 160:
            case 161:
            case 162:
            case 163:
            case 164:
            case 165:
            case 166:
            case 167:
            case 168:
            case 198:
            case 199: {
                buf.append("\t\t#").append(bytes.getIndex() - 1 + bytes.readShort());
                break;
            }
            case 200:
            case 201: {
                buf.append("\t\t#").append(bytes.getIndex() - 1 + bytes.readInt());
                break;
            }
            case 21:
            case 22:
            case 23:
            case 24:
            case 25:
            case 54:
            case 55:
            case 56:
            case 57:
            case 58:
            case 169: {
                int vindex;
                if (Utility.wide) {
                    vindex = bytes.readUnsignedShort();
                    Utility.wide = false;
                }
                else {
                    vindex = bytes.readUnsignedByte();
                }
                buf.append("\t\t%").append(vindex);
                break;
            }
            case 196: {
                Utility.wide = true;
                buf.append("\t(wide)");
                break;
            }
            case 188: {
                buf.append("\t\t<").append(Const.getTypeName(bytes.readByte())).append(">");
                break;
            }
            case 178:
            case 179:
            case 180:
            case 181: {
                final int index = bytes.readUnsignedShort();
                buf.append("\t\t").append(constantPool.constantToString(index, (byte)9)).append(verbose ? (" (" + index + ")") : "");
                break;
            }
            case 187:
            case 192: {
                buf.append("\t");
            }
            case 193: {
                final int index = bytes.readUnsignedShort();
                buf.append("\t<").append(constantPool.constantToString(index, (byte)7)).append(">").append(verbose ? (" (" + index + ")") : "");
                break;
            }
            case 183:
            case 184: {
                final int index = bytes.readUnsignedShort();
                final Constant c = constantPool.getConstant(index);
                buf.append("\t").append(constantPool.constantToString(index, c.getTag())).append(verbose ? (" (" + index + ")") : "");
                break;
            }
            case 182: {
                final int index = bytes.readUnsignedShort();
                buf.append("\t").append(constantPool.constantToString(index, (byte)10)).append(verbose ? (" (" + index + ")") : "");
                break;
            }
            case 185: {
                final int index = bytes.readUnsignedShort();
                final int nargs = bytes.readUnsignedByte();
                buf.append("\t").append(constantPool.constantToString(index, (byte)11)).append(verbose ? (" (" + index + ")\t") : "").append(nargs).append("\t").append(bytes.readUnsignedByte());
                break;
            }
            case 186: {
                final int index = bytes.readUnsignedShort();
                buf.append("\t").append(constantPool.constantToString(index, (byte)18)).append(verbose ? (" (" + index + ")\t") : "").append(bytes.readUnsignedByte()).append(bytes.readUnsignedByte());
                break;
            }
            case 19:
            case 20: {
                final int index = bytes.readUnsignedShort();
                buf.append("\t\t").append(constantPool.constantToString(index, constantPool.getConstant(index).getTag())).append(verbose ? (" (" + index + ")") : "");
                break;
            }
            case 18: {
                final int index = bytes.readUnsignedByte();
                buf.append("\t\t").append(constantPool.constantToString(index, constantPool.getConstant(index).getTag())).append(verbose ? (" (" + index + ")") : "");
                break;
            }
            case 189: {
                final int index = bytes.readUnsignedShort();
                buf.append("\t\t<").append(compactClassName(constantPool.getConstantString(index, (byte)7), false)).append(">").append(verbose ? (" (" + index + ")") : "");
                break;
            }
            case 197: {
                final int index = bytes.readUnsignedShort();
                final int dimensions = bytes.readUnsignedByte();
                buf.append("\t<").append(compactClassName(constantPool.getConstantString(index, (byte)7), false)).append(">\t").append(dimensions).append(verbose ? (" (" + index + ")") : "");
                break;
            }
            case 132: {
                int vindex;
                int constant;
                if (Utility.wide) {
                    vindex = bytes.readUnsignedShort();
                    constant = bytes.readShort();
                    Utility.wide = false;
                }
                else {
                    vindex = bytes.readUnsignedByte();
                    constant = bytes.readByte();
                }
                buf.append("\t\t%").append(vindex).append("\t").append(constant);
                break;
            }
            default: {
                if (Const.getNoOfOperands(opcode) > 0) {
                    for (int k = 0; k < Const.getOperandTypeCount(opcode); ++k) {
                        buf.append("\t\t");
                        switch (Const.getOperandType(opcode, k)) {
                            case 8: {
                                buf.append(bytes.readByte());
                                break;
                            }
                            case 9: {
                                buf.append(bytes.readShort());
                                break;
                            }
                            case 10: {
                                buf.append(bytes.readInt());
                                break;
                            }
                            default: {
                                throw new IllegalStateException("Unreachable default case reached!");
                            }
                        }
                    }
                    break;
                }
                break;
            }
        }
        return buf.toString();
    }
    
    public static String compactClassName(final String str) {
        return compactClassName(str, true);
    }
    
    public static String compactClassName(final String str, final boolean chopit) {
        return compactClassName(str, "java.lang.", chopit);
    }
    
    public static String compactClassName(String str, final String prefix, final boolean chopit) {
        final int len = prefix.length();
        str = pathToPackage(str);
        if (chopit && str.startsWith(prefix) && str.substring(len).indexOf(46) == -1) {
            str = str.substring(len);
        }
        return str;
    }
    
    public static String pathToPackage(final String str) {
        return str.replace('/', '.');
    }
    
    public static String convertString(final String label) {
        final char[] ch = label.toCharArray();
        final StringBuilder buf = new StringBuilder();
        for (final char element : ch) {
            switch (element) {
                case '\n': {
                    buf.append("\\n");
                    break;
                }
                case '\r': {
                    buf.append("\\r");
                    break;
                }
                case '\"': {
                    buf.append("\\\"");
                    break;
                }
                case '\'': {
                    buf.append("\\'");
                    break;
                }
                case '\\': {
                    buf.append("\\\\");
                    break;
                }
                default: {
                    buf.append(element);
                    break;
                }
            }
        }
        return buf.toString();
    }
    
    private static int countBrackets(final String brackets) {
        final char[] chars = brackets.toCharArray();
        int count = 0;
        boolean open = false;
        for (final char c : chars) {
            switch (c) {
                case '[': {
                    if (open) {
                        throw new IllegalArgumentException("Illegally nested brackets:" + brackets);
                    }
                    open = true;
                    break;
                }
                case ']': {
                    if (!open) {
                        throw new IllegalArgumentException("Illegally nested brackets:" + brackets);
                    }
                    open = false;
                    ++count;
                    break;
                }
            }
        }
        if (open) {
            throw new IllegalArgumentException("Illegally nested brackets:" + brackets);
        }
        return count;
    }
    
    public static byte[] decode(final String s, final boolean uncompress) throws IOException {
        byte[] bytes;
        try (JavaReader jr = new JavaReader(new CharArrayReader(s.toCharArray()));
             ByteArrayOutputStream bos = new ByteArrayOutputStream()) {
            int ch;
            while ((ch = jr.read()) >= 0) {
                bos.write(ch);
            }
            bytes = bos.toByteArray();
        }
        if (uncompress) {
            final GZIPInputStream gis = new GZIPInputStream(new ByteArrayInputStream(bytes));
            final byte[] tmp = new byte[bytes.length * 3];
            int count = 0;
            int b;
            while ((b = gis.read()) >= 0) {
                tmp[count++] = (byte)b;
            }
            bytes = Arrays.copyOf(tmp, count);
        }
        return bytes;
    }
    
    public static String encode(byte[] bytes, final boolean compress) throws IOException {
        if (compress) {
            try (ByteArrayOutputStream baos = new ByteArrayOutputStream();
                 GZIPOutputStream gos = new GZIPOutputStream(baos)) {
                gos.write(bytes, 0, bytes.length);
                gos.close();
                bytes = baos.toByteArray();
            }
        }
        final CharArrayWriter caw = new CharArrayWriter();
        try (JavaWriter jw = new JavaWriter(caw)) {
            for (final byte b : bytes) {
                final int in = b & 0xFF;
                jw.write(in);
            }
        }
        return caw.toString();
    }
    
    static boolean equals(final byte[] a, final byte[] b) {
        final int size;
        if ((size = a.length) != b.length) {
            return false;
        }
        for (int i = 0; i < size; ++i) {
            if (a[i] != b[i]) {
                return false;
            }
        }
        return true;
    }
    
    public static String fillup(final String str, final int length, final boolean leftJustify, final char fill) {
        final int len = length - str.length();
        final char[] buf = new char[Math.max(len, 0)];
        Arrays.fill(buf, fill);
        if (leftJustify) {
            return str + new String(buf);
        }
        return new String(buf) + str;
    }
    
    public static String format(final int i, final int length, final boolean leftJustify, final char fill) {
        return fillup(Integer.toString(i), length, leftJustify, fill);
    }
    
    public static String getSignature(String type) {
        final StringBuilder buf = new StringBuilder();
        final char[] chars = type.toCharArray();
        boolean charFound = false;
        boolean delim = false;
        int index = -1;
    Label_0162:
        for (int i = 0; i < chars.length; ++i) {
            switch (chars[i]) {
                case '\t':
                case '\n':
                case '\f':
                case '\r':
                case ' ': {
                    if (charFound) {
                        delim = true;
                        break;
                    }
                    break;
                }
                case '[': {
                    if (!charFound) {
                        throw new IllegalArgumentException("Illegal type: " + type);
                    }
                    index = i;
                    break Label_0162;
                }
                default: {
                    charFound = true;
                    if (!delim) {
                        buf.append(chars[i]);
                        break;
                    }
                    break;
                }
            }
        }
        int brackets = 0;
        if (index > 0) {
            brackets = countBrackets(type.substring(index));
        }
        type = buf.toString();
        buf.setLength();
        for (int j = 0; j < brackets; ++j) {
            buf.append('[');
        }
        boolean found = false;
        for (int k = 4; k <= 12 && !found; ++k) {
            if (Const.getTypeName(k).equals(type)) {
                found = true;
                buf.append(Const.getShortTypeName(k));
            }
        }
        if (!found) {
            buf.append('L').append(type.replace('.', '/')).append(';');
        }
        return buf.toString();
    }
    
    public static boolean isJavaIdentifierPart(final char ch) {
        return (ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z') || (ch >= '0' && ch <= '9') || ch == '_';
    }
    
    public static boolean isSet(final int flag, final int i) {
        return (flag & pow2(i)) != 0x0;
    }
    
    public static String[] methodSignatureArgumentTypes(final String signature) throws ClassFormatException {
        return methodSignatureArgumentTypes(signature, true);
    }
    
    public static String[] methodSignatureArgumentTypes(final String signature, final boolean chopit) throws ClassFormatException {
        final List<String> vec = new ArrayList<String>();
        try {
            int index = signature.indexOf(40) + 1;
            if (index <= 0) {
                throw new ClassFormatException("Invalid method signature: " + signature);
            }
            while (signature.charAt(index) != ')') {
                vec.add(typeSignatureToString(signature.substring(index), chopit));
                index += unwrap(Utility.CONSUMER_CHARS);
            }
        }
        catch (final StringIndexOutOfBoundsException e) {
            throw new ClassFormatException("Invalid method signature: " + signature, e);
        }
        return vec.toArray(ArrayUtils.EMPTY_STRING_ARRAY);
    }
    
    public static String methodSignatureReturnType(final String signature) throws ClassFormatException {
        return methodSignatureReturnType(signature, true);
    }
    
    public static String methodSignatureReturnType(final String signature, final boolean chopit) throws ClassFormatException {
        String type;
        try {
            final int index = signature.lastIndexOf(41) + 1;
            if (index <= 0) {
                throw new ClassFormatException("Invalid method signature: " + signature);
            }
            type = typeSignatureToString(signature.substring(index), chopit);
        }
        catch (final StringIndexOutOfBoundsException e) {
            throw new ClassFormatException("Invalid method signature: " + signature, e);
        }
        return type;
    }
    
    public static String methodSignatureToString(final String signature, final String name, final String access) {
        return methodSignatureToString(signature, name, access, true);
    }
    
    public static String methodSignatureToString(final String signature, final String name, final String access, final boolean chopit) {
        return methodSignatureToString(signature, name, access, chopit, null);
    }
    
    public static String methodSignatureToString(final String signature, final String name, final String access, final boolean chopit, final LocalVariableTable vars) throws ClassFormatException {
        final StringBuilder buf = new StringBuilder("(");
        int varIndex = access.contains("static") ? 0 : 1;
        String type;
        try {
            int index = signature.indexOf(40) + 1;
            if (index <= 0) {
                throw new ClassFormatException("Invalid method signature: " + signature);
            }
            while (signature.charAt(index) != ')') {
                final String paramType = typeSignatureToString(signature.substring(index), chopit);
                buf.append(paramType);
                if (vars != null) {
                    final LocalVariable l = vars.getLocalVariable(varIndex, 0);
                    if (l != null) {
                        buf.append(" ").append(l.getName());
                    }
                }
                else {
                    buf.append(" arg").append(varIndex);
                }
                if ("double".equals(paramType) || "long".equals(paramType)) {
                    varIndex += 2;
                }
                else {
                    ++varIndex;
                }
                buf.append(", ");
                index += unwrap(Utility.CONSUMER_CHARS);
            }
            ++index;
            type = typeSignatureToString(signature.substring(index), chopit);
        }
        catch (final StringIndexOutOfBoundsException e) {
            throw new ClassFormatException("Invalid method signature: " + signature, e);
        }
        if (buf.length() > 1) {
            buf.setLength();
        }
        buf.append(")");
        return access + (access.isEmpty() ? "" : " ") + type + " " + name + buf.toString();
    }
    
    public static String methodTypeToSignature(final String ret, final String[] argv) throws ClassFormatException {
        final StringBuilder buf = new StringBuilder("(");
        if (argv != null) {
            for (final String element : argv) {
                final String str = getSignature(element);
                if (str.endsWith("V")) {
                    throw new ClassFormatException("Invalid type: " + element);
                }
                buf.append(str);
            }
        }
        final String str = getSignature(ret);
        buf.append(")").append(str);
        return buf.toString();
    }
    
    private static int pow2(final int n) {
        return 1 << n;
    }
    
    public static String printArray(final Object[] obj) {
        return printArray(obj, true);
    }
    
    public static String printArray(final Object[] obj, final boolean braces) {
        return printArray(obj, braces, false);
    }
    
    public static String printArray(final Object[] obj, final boolean braces, final boolean quote) {
        if (obj == null) {
            return null;
        }
        final StringBuilder buf = new StringBuilder();
        if (braces) {
            buf.append('{');
        }
        for (int i = 0; i < obj.length; ++i) {
            if (obj[i] != null) {
                buf.append(quote ? "\"" : "").append(obj[i]).append(quote ? "\"" : "");
            }
            else {
                buf.append("null");
            }
            if (i < obj.length - 1) {
                buf.append(", ");
            }
        }
        if (braces) {
            buf.append('}');
        }
        return buf.toString();
    }
    
    public static void printArray(final PrintStream out, final Object[] obj) {
        out.println(printArray(obj, true));
    }
    
    public static void printArray(final PrintWriter out, final Object[] obj) {
        out.println(printArray(obj, true));
    }
    
    public static String replace(String str, final String old, final String new_) {
        try {
            if (str.contains(old)) {
                final StringBuilder buf = new StringBuilder();
                int oldIndex;
                int index;
                for (oldIndex = 0; (index = str.indexOf(old, oldIndex)) != -1; oldIndex = index + old.length()) {
                    buf.append(str, oldIndex, index);
                    buf.append(new_);
                }
                buf.append(str.substring(oldIndex));
                str = buf.toString();
            }
        }
        catch (final StringIndexOutOfBoundsException e) {
            System.err.println(e);
        }
        return str;
    }
    
    public static short searchOpcode(String name) {
        name = name.toLowerCase(Locale.ENGLISH);
        for (short i = 0; i < Const.OPCODE_NAMES_LENGTH; ++i) {
            if (Const.getOpcodeName(i).equals(name)) {
                return i;
            }
        }
        return -1;
    }
    
    public static int setBit(final int flag, final int i) {
        return flag | pow2(i);
    }
    
    public static String signatureToString(final String signature) {
        return signatureToString(signature, true);
    }
    
    public static String signatureToString(final String signature, final boolean chopit) {
        String type = "";
        String typeParams = "";
        int index = 0;
        if (signature.charAt(0) == '<') {
            typeParams = typeParamTypesToString(signature, chopit);
            index += unwrap(Utility.CONSUMER_CHARS);
        }
        if (signature.charAt(index) == '(') {
            type = typeParams + typeSignaturesToString(signature.substring(index), chopit, ')');
            index += unwrap(Utility.CONSUMER_CHARS);
            type += typeSignatureToString(signature.substring(index), chopit);
            index += unwrap(Utility.CONSUMER_CHARS);
            return type;
        }
        type = typeSignatureToString(signature.substring(index), chopit);
        index += unwrap(Utility.CONSUMER_CHARS);
        if (typeParams.isEmpty() && index == signature.length()) {
            return type;
        }
        final StringBuilder typeClass = new StringBuilder(typeParams);
        typeClass.append(" extends ");
        typeClass.append(type);
        if (index < signature.length()) {
            typeClass.append(" implements ");
            typeClass.append(typeSignatureToString(signature.substring(index), chopit));
            index += unwrap(Utility.CONSUMER_CHARS);
        }
        while (index < signature.length()) {
            typeClass.append(", ");
            typeClass.append(typeSignatureToString(signature.substring(index), chopit));
            index += unwrap(Utility.CONSUMER_CHARS);
        }
        return typeClass.toString();
    }
    
    public static String toHexString(final byte[] bytes) {
        final StringBuilder buf = new StringBuilder();
        for (int i = 0; i < bytes.length; ++i) {
            final short b = byteToShort(bytes[i]);
            final String hex = Integer.toHexString(b);
            if (b < 16) {
                buf.append('0');
            }
            buf.append(hex);
            if (i < bytes.length - 1) {
                buf.append(' ');
            }
        }
        return buf.toString();
    }
    
    public static byte typeOfMethodSignature(final String signature) throws ClassFormatException {
        try {
            if (signature.charAt(0) != '(') {
                throw new ClassFormatException("Invalid method signature: " + signature);
            }
            final int index = signature.lastIndexOf(41) + 1;
            return typeOfSignature(signature.substring(index));
        }
        catch (final StringIndexOutOfBoundsException e) {
            throw new ClassFormatException("Invalid method signature: " + signature, e);
        }
    }
    
    public static byte typeOfSignature(final String signature) throws ClassFormatException {
        try {
            switch (signature.charAt(0)) {
                case 'B': {
                    return 8;
                }
                case 'C': {
                    return 5;
                }
                case 'D': {
                    return 7;
                }
                case 'F': {
                    return 6;
                }
                case 'I': {
                    return 10;
                }
                case 'J': {
                    return 11;
                }
                case 'L':
                case 'T': {
                    return 14;
                }
                case '[': {
                    return 13;
                }
                case 'V': {
                    return 12;
                }
                case 'Z': {
                    return 4;
                }
                case 'S': {
                    return 9;
                }
                case '!':
                case '*':
                case '+': {
                    return typeOfSignature(signature.substring(1));
                }
                default: {
                    throw new ClassFormatException("Invalid method signature: " + signature);
                }
            }
        }
        catch (final StringIndexOutOfBoundsException e) {
            throw new ClassFormatException("Invalid method signature: " + signature, e);
        }
    }
    
    private static String typeParamTypesToString(final String signature, final boolean chopit) {
        final StringBuilder typeParams = new StringBuilder("<");
        int index = 1;
        typeParams.append(typeParamTypeToString(signature.substring(index), chopit));
        for (index += unwrap(Utility.CONSUMER_CHARS); signature.charAt(index) != '>'; index += unwrap(Utility.CONSUMER_CHARS)) {
            typeParams.append(", ");
            typeParams.append(typeParamTypeToString(signature.substring(index), chopit));
        }
        wrap(Utility.CONSUMER_CHARS, index + 1);
        return typeParams.append(">").toString();
    }
    
    private static String typeParamTypeToString(final String signature, final boolean chopit) {
        int index = signature.indexOf(58);
        if (index <= 0) {
            throw new ClassFormatException("Invalid type parameter signature: " + signature);
        }
        final StringBuilder typeParam = new StringBuilder(signature.substring(0, index));
        ++index;
        if (signature.charAt(index) != ':') {
            typeParam.append(" extends ");
            typeParam.append(typeSignatureToString(signature.substring(index), chopit));
            index += unwrap(Utility.CONSUMER_CHARS);
        }
        while (signature.charAt(index) == ':') {
            ++index;
            typeParam.append(" & ");
            typeParam.append(typeSignatureToString(signature.substring(index), chopit));
            index += unwrap(Utility.CONSUMER_CHARS);
        }
        wrap(Utility.CONSUMER_CHARS, index);
        return typeParam.toString();
    }
    
    private static String typeSignaturesToString(final String signature, final boolean chopit, final char term) {
        final StringBuilder typeList = new StringBuilder(signature.substring(0, 1));
        int index = 1;
        if (signature.charAt(index) != term) {
            typeList.append(typeSignatureToString(signature.substring(index), chopit));
            index += unwrap(Utility.CONSUMER_CHARS);
        }
        while (signature.charAt(index) != term) {
            typeList.append(", ");
            typeList.append(typeSignatureToString(signature.substring(index), chopit));
            index += unwrap(Utility.CONSUMER_CHARS);
        }
        wrap(Utility.CONSUMER_CHARS, index + 1);
        return typeList.append(term).toString();
    }
    
    public static String typeSignatureToString(final String signature, final boolean chopit) throws ClassFormatException {
        wrap(Utility.CONSUMER_CHARS, 1);
        try {
            switch (signature.charAt(0)) {
                case 'B': {
                    return "byte";
                }
                case 'C': {
                    return "char";
                }
                case 'D': {
                    return "double";
                }
                case 'F': {
                    return "float";
                }
                case 'I': {
                    return "int";
                }
                case 'J': {
                    return "long";
                }
                case 'T': {
                    final int index = signature.indexOf(59);
                    if (index < 0) {
                        throw new ClassFormatException("Invalid type variable signature: " + signature);
                    }
                    wrap(Utility.CONSUMER_CHARS, index + 1);
                    return compactClassName(signature.substring(1, index), chopit);
                }
                case 'L': {
                    int fromIndex = signature.indexOf(60);
                    if (fromIndex < 0) {
                        fromIndex = 0;
                    }
                    else {
                        fromIndex = signature.indexOf(62, fromIndex);
                        if (fromIndex < 0) {
                            throw new ClassFormatException("Invalid signature: " + signature);
                        }
                    }
                    final int index2 = signature.indexOf(59, fromIndex);
                    if (index2 < 0) {
                        throw new ClassFormatException("Invalid signature: " + signature);
                    }
                    final int bracketIndex = signature.substring(0, index2).indexOf(60);
                    if (bracketIndex < 0) {
                        wrap(Utility.CONSUMER_CHARS, index2 + 1);
                        return compactClassName(signature.substring(1, index2), chopit);
                    }
                    fromIndex = signature.indexOf(59);
                    if (fromIndex < 0) {
                        throw new ClassFormatException("Invalid signature: " + signature);
                    }
                    if (fromIndex < bracketIndex) {
                        wrap(Utility.CONSUMER_CHARS, fromIndex + 1);
                        return compactClassName(signature.substring(1, fromIndex), chopit);
                    }
                    final StringBuilder type = new StringBuilder(compactClassName(signature.substring(1, bracketIndex), chopit)).append("<");
                    int consumedChars = bracketIndex + 1;
                    if (signature.charAt(consumedChars) == '+') {
                        type.append("? extends ");
                        ++consumedChars;
                    }
                    else if (signature.charAt(consumedChars) == '-') {
                        type.append("? super ");
                        ++consumedChars;
                    }
                    if (signature.charAt(consumedChars) == '*') {
                        type.append("?");
                        ++consumedChars;
                    }
                    else {
                        type.append(typeSignatureToString(signature.substring(consumedChars), chopit));
                        consumedChars += unwrap(Utility.CONSUMER_CHARS);
                        wrap(Utility.CONSUMER_CHARS, consumedChars);
                    }
                    while (signature.charAt(consumedChars) != '>') {
                        type.append(", ");
                        if (signature.charAt(consumedChars) == '+') {
                            type.append("? extends ");
                            ++consumedChars;
                        }
                        else if (signature.charAt(consumedChars) == '-') {
                            type.append("? super ");
                            ++consumedChars;
                        }
                        if (signature.charAt(consumedChars) == '*') {
                            type.append("?");
                            ++consumedChars;
                        }
                        else {
                            type.append(typeSignatureToString(signature.substring(consumedChars), chopit));
                            consumedChars += unwrap(Utility.CONSUMER_CHARS);
                            wrap(Utility.CONSUMER_CHARS, consumedChars);
                        }
                    }
                    ++consumedChars;
                    type.append(">");
                    if (signature.charAt(consumedChars) == '.') {
                        type.append(".");
                        type.append(typeSignatureToString("L" + signature.substring(consumedChars + 1), chopit));
                        consumedChars += unwrap(Utility.CONSUMER_CHARS);
                        wrap(Utility.CONSUMER_CHARS, consumedChars);
                        return type.toString();
                    }
                    if (signature.charAt(consumedChars) != ';') {
                        throw new ClassFormatException("Invalid signature: " + signature);
                    }
                    wrap(Utility.CONSUMER_CHARS, consumedChars + 1);
                    return type.toString();
                }
                case 'S': {
                    return "short";
                }
                case 'Z': {
                    return "boolean";
                }
                case '[': {
                    final StringBuilder brackets = new StringBuilder();
                    int n;
                    for (n = 0; signature.charAt(n) == '['; ++n) {
                        brackets.append("[]");
                    }
                    final int consumedChars2 = n;
                    final String type2 = typeSignatureToString(signature.substring(n), chopit);
                    final int temp = unwrap(Utility.CONSUMER_CHARS) + consumedChars2;
                    wrap(Utility.CONSUMER_CHARS, temp);
                    return type2 + brackets.toString();
                }
                case 'V': {
                    return "void";
                }
                default: {
                    throw new ClassFormatException("Invalid signature: `" + signature + "'");
                }
            }
        }
        catch (final StringIndexOutOfBoundsException e) {
            throw new ClassFormatException("Invalid signature: " + signature, e);
        }
    }
    
    private static int unwrap(final ThreadLocal<Integer> tl) {
        return tl.get();
    }
    
    private static void wrap(final ThreadLocal<Integer> tl, final int value) {
        tl.set(value);
    }
    
    static {
        CONSUMER_CHARS = ThreadLocal.withInitial(() -> 0);
        CHAR_MAP = new int[48];
        MAP_CHAR = new int[256];
        int j = 0;
        for (int i = 65; i <= 90; ++i) {
            Utility.CHAR_MAP[j] = i;
            Utility.MAP_CHAR[i] = j;
            ++j;
        }
        for (int i = 103; i <= 122; ++i) {
            Utility.CHAR_MAP[j] = i;
            Utility.MAP_CHAR[i] = j;
            ++j;
        }
        Utility.CHAR_MAP[j] = 36;
        Utility.MAP_CHAR[36] = j;
        ++j;
        Utility.CHAR_MAP[j] = 95;
        Utility.MAP_CHAR[95] = j;
    }
    
    private static class JavaReader extends FilterReader
    {
        public JavaReader(final Reader in) {
            super(in);
        }
        
        @Override
        public int read() throws IOException {
            final int b = this.in.read();
            if (b != 36) {
                return b;
            }
            final int i = this.in.read();
            if (i < 0) {
                return -1;
            }
            if ((i < 48 || i > 57) && (i < 97 || i > 102)) {
                return Utility.MAP_CHAR[i];
            }
            final int j = this.in.read();
            if (j < 0) {
                return -1;
            }
            final char[] tmp = { (char)i, (char)j };
            return Integer.parseInt(new String(tmp), 16);
        }
        
        @Override
        public int read(final char[] cbuf, final int off, final int len) throws IOException {
            for (int i = 0; i < len; ++i) {
                cbuf[off + i] = (char)this.read();
            }
            return len;
        }
    }
    
    private static class JavaWriter extends FilterWriter
    {
        public JavaWriter(final Writer out) {
            super(out);
        }
        
        @Override
        public void write(final char[] cbuf, final int off, final int len) throws IOException {
            for (int i = 0; i < len; ++i) {
                this.write(cbuf[off + i]);
            }
        }
        
        @Override
        public void write(final int b) throws IOException {
            if (Utility.isJavaIdentifierPart((char)b) && b != 36) {
                this.out.write(b);
            }
            else {
                this.out.write(36);
                if (b >= 0 && b < 48) {
                    this.out.write(Utility.CHAR_MAP[b]);
                }
                else {
                    final char[] tmp = Integer.toHexString(b).toCharArray();
                    if (tmp.length == 1) {
                        this.out.write(48);
                        this.out.write(tmp[0]);
                    }
                    else {
                        this.out.write(tmp[0]);
                        this.out.write(tmp[1]);
                    }
                }
            }
        }
        
        @Override
        public void write(final String str, final int off, final int len) throws IOException {
            this.write(str.toCharArray(), off, len);
        }
    }
}
